//this generates the number of XOR operations for each element of the field
#include "header.h"

//These generates the XOR tables
//==============================================================================
//Performs GF(2^n) field multiplication
unsigned int field_multiply(unsigned int a1, unsigned int a2, const unsigned int poly, int fieldexp){
    unsigned int c2=0, poly_;

    if(fieldexp==2) poly_ = poly & 0x3;
    else if(fieldexp==3) poly_ = poly & 0x7;
    else if(fieldexp==4) poly_ = poly & 0xf;
    else if(fieldexp==5) poly_ = poly & 0x1f;
    else if(fieldexp==6) poly_ = poly & 0x3f;
    else if(fieldexp==7) poly_ = poly & 0x7f;
    else if(fieldexp==8) poly_ = poly & 0xff;
    else if(fieldexp==9) poly_ = poly & 0x1ff;
    else poly_ = poly & 0x3ff;

    while(a2!=0){
        if(a2 & 1) c2 ^= a1;
        a2 >>= 1;
        a1 <<= 1;
        if(a1 & (1 << fieldexp)) a1 = (a1 & ((1 << fieldexp) - 1 )) ^ poly_;
    }
    return(c2);
}

//Produces binary representation of a number
//express w as array of binary in b, storing in lsb order
void binary_rep(const unsigned int w, int *b, int fieldexp){
    int i;
    for(i=0; i<fieldexp; i++) b[i] = (w >> i) & 0x01;
}

//Returns the #XORs for multiplication of a field element x with each element
//output the XOR representation in M
int xor_count(const unsigned int a, const unsigned int poly, int fieldexp, int *M){
    int i, l, h, p[fieldexp], exp[fieldexp][fieldexp], m[fieldexp], q[fieldexp];
    int c=0, counter=0;
    unsigned int Q;

    binary_rep(poly, p, fieldexp);

    //initialise exp[][] to be all zeroes
    for(l=0; l<fieldexp; l++)
        for(h=0; h<fieldexp; h++) exp[l][h] = 0;

    //convert a into binary
    binary_rep(a, m, fieldexp);

    //store in exp the bits that represents the multiplication
    for(i=0; i<fieldexp; i++){
        if(m[i]==1){                             //i denotes bit position of a
            for(h=0; h<fieldexp; h++){           //h denotes bit position of x
                Q = field_multiply((1<<i), (1<<h), poly, fieldexp);
                binary_rep(Q, q, fieldexp);
                for(l=0; l<fieldexp; l++)  //goes through each bit position of Q
                    if(q[l]==1) exp[fieldexp-1 - l][fieldexp-1 - h] = exp[fieldexp-1 - l][fieldexp-1 - h]^1;
            }
        }
    }

    for(i=0; i<fieldexp;i++){
    	M[i] =0;
    	for(h=0; h<fieldexp; h++){
    		M[i] ^= (exp[i][fieldexp-1-h]<<h);
    	}
    }

    for(l=0; l<fieldexp; l++){
        for(h=0; h<fieldexp; h++)
            if(exp[fieldexp-l-1][h]==1) c = h;
        for(h=0; h<fieldexp; h++)
            if((exp[fieldexp-l-1][h]==1) && (h!=c)) counter = counter + 1;
    }
    return counter;
}

//==========================================================================================================================================================================

void compute_multi_rep(int Multi_rep[][fieldsize]){

    int order = 1<< fieldsize;

	cout<<"Poly field = "<<std::hex<<"0x"<<fieldpoly<<std::dec<<endl;

    int M[fieldsize];

        for (int i = 1; i< order; i++){

            xor_count(i, fieldpoly, fieldsize, M);
            for (int j=0; j<fieldsize;j++){
                Multi_rep[i][j] = M[j];
            }
        }

    return;
}
