#include "header.h"

int best_seq[64][2]={};
int current_seq[64][2]={};

//recurrence function for GJ_algo
int GJ_recur(int GJ_min_xor, int current_xor, int *M, int* row_order, int row_step){
//while the current number of steps is not equal to current best, continue to simplify
    //dont need to check "current_xor<GJ_min_xor" first, at the end if fail, do not call this function again.

        //compute the direct XOR count value
        int perm_id = 0;
        for (int i=0; i<fieldsize; i++){
            perm_id += __builtin_popcount(M[i]) - 1;
        }
        //if perm_id=0, we have a row permutation of an identity matrix
        if (perm_id == 0){
            for (int i= 0; i<current_xor; i++){
                best_seq[i][0] = current_seq[i][0];
                best_seq[i][1] = current_seq[i][1];
            }

            //also copy the current state in the next few values
            for (int i= current_xor; i<current_xor+fieldsize; i++){
                best_seq[i][0] = M[i-current_xor];
                best_seq[i][1] = -1;
            }

            return current_xor;
        }

        //else proceed on with the next row_step then call the function if "temp_current_xor<GJ_min_xor"
        //get the index of the row that will be used in this recursion
        int row = row_order[row_step];
        int temp_current_xor;
        int temp_M[fieldsize];
        int fixed_pivot_xor;
        //try all possible pivot position
        for (int pivot_pos = 0; pivot_pos < fieldsize; pivot_pos++){
            if (((M[row]>>((fieldsize-1)-pivot_pos))&0x1)==0) continue;     //not a valid pivot_pos
            //after fixing a pivot position, check through all other rows
            temp_current_xor = current_xor;
            //copy a temp matrix state
            for (int i=0; i< fieldsize ;i++) temp_M[i] = M[i];
            //apply GJ eli to the temp_M
            for (int other_row=0; other_row<fieldsize; other_row++){
                if (other_row==row) continue;   //do not XOR the original row
                if (((temp_M[other_row]>>((fieldsize-1)-pivot_pos))&0x1)==1){
                    temp_M[other_row] ^= temp_M[row];
                    current_seq[temp_current_xor][0]=other_row;
                    current_seq[temp_current_xor][1]=row;
                    temp_current_xor++;
                }
            }
            //after applying XOR to all other rows whenever necessary, check if the XOR count exceed
            if (temp_current_xor >= GJ_min_xor) continue;

            //if not, this row_step is done we move on to the next row.
            fixed_pivot_xor = GJ_recur(GJ_min_xor, temp_current_xor, temp_M, row_order, row_step+1);
            if (fixed_pivot_xor< GJ_min_xor) GJ_min_xor = fixed_pivot_xor;

        }


    return GJ_min_xor;
}

//input array of size fieldsize, with M[0] the first row of the matrix
int GJ_algo_DM(DM_VERTEX V, vector <DM_VERTEX> &Stage_state, int transpose, bool printscreen ){
    int M[fieldsize];
    if (transpose) V = DM_transpose(V);
    for (int i=0; i<fieldsize; i++){
        M[i] = V.state[i];
    }

    //compute the number of XOR for the multiplication matrix
    int GJ_min_xor = 0;
    for (int i=0; i<fieldsize; i++){
        GJ_min_xor += __builtin_popcount(M[i]) - 1;
    }
    if (printscreen){
        cout<<"direct counting the number of XOR in multiplication matrix = "<<GJ_min_xor<<endl;
    }

    //to prevent error when optimal is same as naive counting
    GJ_min_xor=1000;

    int current_xor, fixed_row_order_xor;
    int row_order[fieldsize];
    //max for 8x8 matrix
    for (int i=0; i<fieldsize; i++){
        row_order[i]=i;
    }

    //define a row order, apply
    do{
        current_xor=0;
        fixed_row_order_xor = GJ_recur(GJ_min_xor, current_xor, M, row_order, 0);
        if (fixed_row_order_xor < GJ_min_xor) GJ_min_xor = fixed_row_order_xor;
    }
    while ( std::next_permutation(row_order,row_order+fieldsize) );


    //after trying all possible row_order, what is left is the optimal GJ result in best_seq
    if (transpose){
        // V.initialize_DM_VERTEX();

        for (int i=0; i<fieldsize;i++){
            V.state[i] = (1<<(fieldsize-1))>>i;
        }

        Stage_state.clear();
        Stage_state.push_back(V);
        //do the seq in construction order (reverse) and swap the XOR_pos and output_pos
        for (int xor_seq=0; xor_seq<GJ_min_xor; xor_seq++){
            V = XOR_rows(V, best_seq[xor_seq][1], best_seq[xor_seq][0]);
            Stage_state.push_back(V);
        }
    }
    else{
        //update the stage state
        Stage_state.clear();
        Stage_state.insert(Stage_state.begin(),V);
        //the seq are stored in order to reduce V to Id matrix
        for (int xor_seq=0; xor_seq<GJ_min_xor; xor_seq++){
            V = XOR_rows(V, best_seq[xor_seq][0], best_seq[xor_seq][1]);
            Stage_state.insert(Stage_state.begin(),V);
        }
    }


    if (printscreen){
        if (transpose){
            int val = 1<<(fieldsize-1);
            for (int i=GJ_min_xor; i<GJ_min_xor+fieldsize; i++){
                cout<<"M["<<i-GJ_min_xor<<"] = "<<val<<endl;
                val = val>>1;
            }
            for (int i=0; i<GJ_min_xor; i++){
                cout<<"Row "<<best_seq[i][1]<<" ^= Row "<<best_seq[i][0]<<endl;
            }
        }
        else {
            for (int i=GJ_min_xor; i<GJ_min_xor+fieldsize; i++){
                cout<<"M["<<i-GJ_min_xor<<"] = "<<best_seq[i][0]<<endl;
            }
            for (int i=GJ_min_xor-1; i>=0; i--){
                cout<<"Row "<<best_seq[i][0]<<" ^= Row "<<best_seq[i][1]<<endl;
            }
        }

    }

    if (printscreen){
        cout<<"by Gauss-Jordan algo, the number of XOR needed = "<<GJ_min_xor<<endl<<endl;
    }

    return GJ_min_xor;
}
