#include <set>
#include <omp.h>
#include "header.h"

/*
**
*/
void all_paths_given_length_rec(DM_VERTEX V_root,
                                DM_VERTEX V_target,
                                unsigned int length,
                                vector<DM_VERTEX>& current_path,
                                vector< vector<DM_VERTEX> >& paths) {

    vector<DM_VERTEX> sons = getListFromRoot(V_root, 1);
    for(auto son : sons) {

        /* Extend the current path by this node */
        current_path.push_back(son);


        if(EqualDM_VERTEX()(V_target, son)) {
            paths.push_back(current_path);
        } else {
            if(length>1) {
                all_paths_given_length_rec(son, V_target, length-1, current_path, paths);
            }
        }

        /* Remove the current node the from the stack */
        current_path.pop_back();

    }/*v*/


}
/*
**
*/
void all_paths_between_nodes(vector <DM_VERTEX> Stage_state,
                             unsigned int start,
                             unsigned int end,
                             vector< vector<DM_VERTEX> >& result) {

    DM_VERTEX V_root=Stage_state[start];
    DM_VERTEX V_target=Stage_state[end];
    unsigned int length = end-start;

    vector<DM_VERTEX> current_path;
    current_path.push_back(V_root);

    vector< vector<DM_VERTEX> > paths;
    all_paths_given_length_rec(V_root, V_target, length, current_path, paths);

    for(auto p: paths) {
        for(unsigned int i=start; i<=end; ++i) Stage_state[i]=p[i-start];
        result.push_back(Stage_state);
    }/*p*/

}

//input the list of intermediate states; the maximum step to grow; whether to print the details
size_t chain_MITM_recursive(vector <DM_VERTEX> Stage_state,
                            unsigned int start,
                            unsigned int end,
                            unsigned int MITM_size,
                            vector<DM_VERTEX>& Stage_state_optimized,
                            int level,
                            int print_details) {

    printf("-------------------------------------------------\n");
    printf("[-] Enters chain_MITM_recursive: [%d-%d]\n", start, end);
    unsigned int length=end-start;

    /* If we have divided in */
    if(length==0) return Stage_state.size()-1;
    if(length<=MITM_size) {

        if(print_details) printf("-- length=%d for [%d-%d]\n", length, start, end);

        // vector<DM_VERTEX>().swap(Stage_state_optimized);
        MITM_search_shell(Stage_state[start], Stage_state[end], Stage_state_optimized, MITM_size, 1);

        if(print_details) printf("--> Stage_state_optimized size=%ld\n", Stage_state_optimized.size()-1);

        return Stage_state_optimized.size()-1;
    }

    /* In the general case, we still need to split in parts smaller than MITM_size */
    unsigned int a;
    unsigned int min=length;

    // printf("[-] FIX HERE: restore 3!\n");
    for(unsigned int mitm_size=MITM_size; mitm_size>=3/*MITM_size*/; mitm_size--) {
        for(a=0; a<=(length-mitm_size); ++a) {

            /* Consider a steps, then mitm_size steps, then length-a-mitm_size steps */

            vector<DM_VERTEX> tmp_first;
            vector<DM_VERTEX> tmp_second;
            vector<DM_VERTEX> tmp_third;
            bool improved_first=false;
            bool improved_second=false;
            bool improved_third=false;

            // int left=length-a-mitm_size;

            if(print_details) {
                printf("[%d-%d] [%d-%d] [%d-%d]\n",
                        start, start+a,
                        start+a, start+a+mitm_size,
                        start+a+mitm_size, end
                );
            }

            /* First block: 0 to a */
            size_t res_first = chain_MITM_recursive(Stage_state, start, start+a, mitm_size, tmp_first, level+1, print_details);

            if(res_first && res_first<a) improved_first=true;
            else res_first=a;


            /* Second block: a to a+mitm_size */
            size_t res_second = chain_MITM_recursive(Stage_state, start+a, start+a+mitm_size, mitm_size, tmp_second, level+1, print_details);

            if(res_second && res_second<mitm_size) improved_second=true;
            else res_second=mitm_size;

            size_t res_third;

            /* Third block: a+mitm_size to length */
            if((start+a+mitm_size)<end) {

                res_third = chain_MITM_recursive(Stage_state, start+a+mitm_size, end, mitm_size, tmp_third, level+1, print_details);

                if(res_third && res_third<(length-a-mitm_size)) improved_third=true;
                else res_third=length-a-mitm_size;

            } else {
                res_third=0;
            }

            if(improved_first || improved_second || improved_third) {
                size_t sum=res_first+res_second+res_third;

                if(print_details) printf("! improved (level=%d): length=%d --> sum=%ld (%ld+%ld+%ld) -- a=%d\n", level, length, sum, res_first, res_second, res_third, a);

                if(sum<min) {
                    min=sum;

                    /* Construct the new optimized sequence */
                    vector<DM_VERTEX>().swap(Stage_state_optimized);

                    if(improved_first)  Stage_state_optimized.insert(Stage_state_optimized.end(), tmp_first.begin(), tmp_first.end());
                    else                for(unsigned int i=start; i<start+a; ++i) Stage_state_optimized.push_back(Stage_state[i]);

                    if(improved_second) Stage_state_optimized.insert(Stage_state_optimized.end(), tmp_second.begin(), tmp_second.end());
                    else                for(unsigned int i=start+a; i<start+a+mitm_size; ++i) Stage_state_optimized.push_back(Stage_state[i]);

                    if(improved_third)  Stage_state_optimized.insert(Stage_state_optimized.end(), tmp_third.begin(), tmp_third.end());
                    else                for(unsigned int i=start+a+mitm_size; i<=end; ++i) Stage_state_optimized.push_back(Stage_state[i]);

                    // for(auto v: Stage_state) printVertex(v, 0); printf("\n");
                    // printf("-----------------------------------------\n");
                    // printf("tmp_first:\n"); for(auto v: tmp_first)  printVertex(v, 0); printf("\n");
                    // printf("tmp_second:\n");for(auto v: tmp_second) printVertex(v, 0); printf("\n");
                    // printf("tmp_third:\n"); for(auto v: tmp_third)  printVertex(v, 0); printf("\n");

                    // printf("|Stage_state_optimized|=%ld\n", Stage_state_optimized.size());
                    // for(auto v: Stage_state_optimized) printVertex(v, 0);

                    auto last = unique(Stage_state_optimized.begin(), Stage_state_optimized.end(), EqualDM_VERTEX());
                    Stage_state_optimized.erase(last, Stage_state_optimized.end());

                    // printf("-----------------------------------------\n");
                    // printf("[%d-%d] [%d-%d] [%d-%d]\n",
                    //         start, start+a,
                    //         start+a, start+a+mitm_size,
                    //         start+a+mitm_size, end
                    // );
                    // printf("|Stage_state_optimized|=%ld (uniqued)\n", Stage_state_optimized.size());
                    // for(auto v: Stage_state_optimized) printVertex(v, 0);
                    // // for(auto v: Stage_state_optimized) printVertex_inst(v);

                }

            }/*if*/

        }/*a*/

    }/*mitm_size*/

    printf("[-] Leaves chain_MITM_recursive: [%d–%d] with min=%d\n", start, end, min);
    return min;
}




//input the list of intermediate states; the maximum step to grow; whether to print the details
vector<DM_VERTEX> chain_MITM(vector <DM_VERTEX> Stage_state, int MITM_size, int print_details){

    ///////////////////////// RECURSIVE OPTIMIZATION //////////////////////////////////
    for(size_t x=0; x<Stage_state.size(); ++x) sort_V_SortNetwork(Stage_state[x]);

    vector<DM_VERTEX> Stage_state_optimized;
    size_t res = chain_MITM_recursive(Stage_state, 0, Stage_state.size()-1, MITM_size, Stage_state_optimized, 0, print_details);
    if(print_details) printf("-- chain_MITM returned: %ld (size=%ld)\n", res, Stage_state_optimized.size());

    if(Stage_state_optimized.size()) return Stage_state_optimized;
    else return Stage_state;
    ///////////////////////////////////////////////////////////////////////////////////






    for(size_t x=0; x<Stage_state.size(); ++x) sort_V_SortNetwork(Stage_state[x]);

    //get the existing seq size
    int step_size = Stage_state.size()-1;

    if (print_details) cout<<"Given path has "<<step_size<<" steps."<<endl;
    if (print_details) cout<<"Finding improvements in blocks of "<<MITM_size+1<<" steps."<<endl;

    //the number of max step MITM search+1 because if search is empty, max step+1 is the optimal solution.
    int ROUNDS = step_size/(MITM_size+1);
    int minimal_steps = 0;
    // vector< vector<int> > Seq;
    vector <DM_VERTEX> NEW_Stage_state, temp_stage_state;
    //add the initial state
    NEW_Stage_state.push_back(Stage_state.at(0));

    for (int i=0; i<ROUNDS; i++) {

        if (print_details) cout << "ROUND " << i << endl;
        // Seq.clear();
        temp_stage_state.clear();

        //currently the seq is not in use besides getting the size
        /*Seq =*/ MITM_search_shell(Stage_state.at(i*(MITM_size+1)), Stage_state.at((i+1)*(MITM_size+1)),temp_stage_state,MITM_size,1);
        // Seq = MITM_search_hashTable(Stage_state.at(i*(MITM_size+1)), Stage_state.at((i+1)*(MITM_size+1)),temp_stage_state,MITM_size,1);

        // printf("seq:%ld - temp:%ld\n", Seq.size(), temp_stage_state.size());

        if (temp_stage_state.size()==0) {
            if (print_details) cout<<"cant improve this round..., number of steps = "<<MITM_size+1<<endl;
            minimal_steps += MITM_size+1;
            //copy the original path, start from j=1, initial state already taken.
            for (int j=1; j<MITM_size+1;j++){
                NEW_Stage_state.push_back(Stage_state.at(i*(MITM_size+1)+j));
            }

        } else{
            if (print_details) cout<<"IMPROVED path, number of steps = "<<(temp_stage_state.size()-1)<<endl;
            minimal_steps += temp_stage_state.size()-1;
            //copy the new path, start from j=1, initial state already taken.
            for (unsigned int j=1; j<temp_stage_state.size()-1;j++){
                NEW_Stage_state.push_back(temp_stage_state.at(j));
            }
        }

        //copy the next round initial state.
        NEW_Stage_state.push_back(Stage_state.at((i+1)*(MITM_size+1)));

        // printf("End of ROUND %d (temp_stage_state):\n", i);
        // for(int x=0; x<temp_stage_state.size(); ++x) printVertex(temp_stage_state[x], 0);
        // printf("\n");


    }/*i*/

    int remaining_steps = step_size-(ROUNDS*(MITM_size+1));
    if (remaining_steps>0) {
        if (print_details) cout << "There are " << remaining_steps << " steps remaining." << endl;

        temp_stage_state.clear();
        MITM_search_shell(Stage_state.at(ROUNDS*(MITM_size+1)), Stage_state.back(),temp_stage_state,remaining_steps-1,1);

        if (temp_stage_state.size()==0){
            if (print_details) cout<<"cant improve remaining steps..., number of steps = "<<remaining_steps<<endl;
            minimal_steps += remaining_steps;
            //copy the original path, start from j=1, initial state already taken.
            for (int j=1; j<remaining_steps;j++){
                NEW_Stage_state.push_back(Stage_state.at(ROUNDS*(MITM_size+1)+j));
            }
        }
        else{
            if (print_details) cout<<"IMPROVED path, number of steps = "<<(temp_stage_state.size()-1)<<endl;
            minimal_steps += temp_stage_state.size()-1;
            //copy the new path, start from j=1, initial state already taken.
            for (unsigned int j=1; j<temp_stage_state.size()-1;j++){
                NEW_Stage_state.push_back(temp_stage_state.at(j));
            }
        }
        //add the final state
        NEW_Stage_state.push_back(Stage_state.back());


        // printf("End of FINAL ROUND:\n");
        // for(int x=0; x<NEW_Stage_state.size(); ++x) printVertex(NEW_Stage_state[x], 0);
        // printf("\n");


    }
    else {
        if (print_details) cout<<"No remaining steps"<<endl;
    }
    if (print_details) cout<<"Done, minimal steps = "<<minimal_steps<<endl;

    // for(int x=0; x<NEW_Stage_state.size(); ++x) {
    //     printVertex(NEW_Stage_state[x], 0);
    // }

    return NEW_Stage_state;
}












int get_XORcount(DM_VERTEX V){
    int counter=0;
    for (int i=0; i< fieldsize;i++){
        counter += (__builtin_popcount(V.state[i]) - 1);
    }
    return counter;
}


int main(int argc, char **argv) {

    clock_t start, stop;
    double t = 0.0;

    /*
    ** Tune OpenMP for parallel computations
    */
    omp_set_dynamic(false);
    omp_set_num_threads(NUM_THREADS);


    /* Start timer */
    start = clock();

    DM_VERTEX V_s,V_e;
    for (int i=0; i<fieldsize;i++){
        V_s.state[i] = (1<<(fieldsize-1))>>i;
    }

    int Multi_rep[1<<fieldsize][fieldsize];
    compute_multi_rep(Multi_rep);

    if(argc!=4) {
        cout << "Usage: " << argv[0] << " <first coeff> <last coeff> <max MITM step>" << endl;
        return 0;
    }

    int input_ele, last_ele;

    /* Get start/end element */
    input_ele=atoi(argv[1]);
    last_ele=atoi(argv[2]);

    /* Get the maximal level for the MITM graph search */
    unsigned int MTM_max_step=atoi(argv[3]);

    if (last_ele>=(1<<fieldsize)) last_ele= (1<<fieldsize)-1;

    ofstream All_XOR_report, MITM_XOR_report, Minimal_Seq_report;
    stringstream AXname,MXname,MSname;
    AXname<<"DM_"<<fieldsize<<"-bit_0x"<<hex<<fieldpoly<<dec<<"_All_XORcount_report_from_"<<input_ele<<"_to_"<<last_ele<<"_MTM"<<MTM_max_step<<".txt";
    MXname<<"DM_"<<fieldsize<<"-bit_0x"<<hex<<fieldpoly<<dec<<"_MITM_XORcount_report_from_"<<input_ele<<"_to_"<<last_ele<<"_MTM"<<MTM_max_step<<".txt";
    MSname<<"DM_"<<fieldsize<<"-bit_0x"<<hex<<fieldpoly<<dec<<"_Minimal_XORseq_report_from_"<<input_ele<<"_to_"<<last_ele<<"_MTM"<<MTM_max_step<<".txt";

    All_XOR_report.open(AXname.str().c_str());
    MITM_XOR_report.open(MXname.str().c_str());
    Minimal_Seq_report.open(MSname.str().c_str());

    All_XOR_report<<"Ele  hex  Ori  GJm  MTM  Compare\n";
    int Ori_best=0, MTM_best=0, GJm_best=0, TOTAL_XOR=0;
    MITM_XOR_report<<"int XORcount_"<<fieldsize<<"_0x"<<hex<<fieldpoly<<dec<<"[]={\n"<<" -1,  0, ";

    /*
    ** Precompute the graph rooted on the Identity
    */
    MITM_precompute(MTM_max_step, 0);

    /*
    ** For all the requested elements
    */
    for (int ele_i=input_ele; ele_i<(last_ele+1); ele_i++){

        for (int i=0; i<fieldsize;i++){
            V_e.state[i] = Multi_rep[ele_i][i];
        }


        unsigned int OriXOR = get_XORcount(V_e);
        cout<<"======================================="<<endl;
        cout<<"Element "<<ele_i<<" has "<<OriXOR<<" XORs"<<endl;
        All_XOR_report<<setw(5)<<left<<ele_i<<setw(5)<<hex<<ele_i<<dec<<setw(5)<<OriXOR;
        Minimal_Seq_report<<"=========================================\n";
        Minimal_Seq_report<<"Element "<<ele_i<<" (0x"<<hex<<setw(2)<<ele_i<<dec<<")";


        vector <DM_VERTEX> Stage_state;
        vector <DM_VERTEX> Stage_stateT;

        cout<<"Try MITM algo..."<<endl;

        Stage_state.clear();
        MITM_search_shell(V_s, V_e, Stage_state, MTM_max_step, 1);

        unsigned int GJ_XOR;
        unsigned int GJT_XOR;
        if (Stage_state.size()==0){

            cout << "No solution within " << MTM_max_step << " steps..." << endl;

            cout << "Computing Gauss-Jordan algo..." << endl;
            GJ_XOR = GJ_algo_DM(V_e, Stage_state, 0, 1);

            cout << "Computing for its transpose..." << endl;
            GJT_XOR = GJ_algo_DM(V_e, Stage_stateT, 1, 1);

            if (GJT_XOR < GJ_XOR){
                cout << "Transpose is better..." << endl;
                GJ_XOR = GJT_XOR;
                Stage_state = Stage_stateT;
            } else {
                cout << "Default is better..." << endl;
            }

            All_XOR_report << setw(5) << GJ_XOR;
            if (GJ_XOR>(MTM_max_step+1)){

                /* Keep applying the chain MITM as long as we are decreasing the length of the implementation */
                size_t length=Stage_state.size();
                while(1) {
                    cout << "Apply MITM technique..." << endl;
                    Stage_state = chain_MITM(Stage_state, MTM_max_step, 1);

                    for(auto v : Stage_state) printVertex(v, 0);
                    // for(auto v: Stage_state) printVertex_inst(v);

                    if(Stage_state.size()<length) length=Stage_state.size();
                    else break;
                }


                /*
                ** If the final length is <=MITM+1, then the implementation is optimal
                ** If it is exactly MITM+2, we try to swap two nodes
                */
                cout << length << endl;
                if((length-1)>=(MTM_max_step+2)) {

                    vector< vector<DM_VERTEX> > paths;
                    all_paths_between_nodes(Stage_state, 0, 3, paths);

                    for(auto p: paths) {

                        cout << "Apply MITM technique..." << endl;
                        for(auto v : p) printVertex(v, 0);

                        Stage_state = chain_MITM(p, MTM_max_step, 1);

                        for(auto v : Stage_state) printVertex(v, 0);

                        if(Stage_state.size()<length) length=Stage_state.size();

                    }/*p*/
                    cout << length << endl;

                }


            } else {
                cout << "Gauss found optimal solution..." << endl;
            }

        } else {
            cout << "Direct MITM has solution!" << endl;
            GJ_XOR=99;
            All_XOR_report << setw(5) << "--";

            // for(auto v : Stage_state) printVertex(v, 0);
            // for(auto v: Stage_state) printVertex_inst(v);

        }
        unsigned int MTMXOR = Stage_state.size()-1;
        TOTAL_XOR += MTMXOR;

        /* Output implementation */
        // for(auto v: Stage_state) printVertex_inst(v);



        /* Write in files */
        All_XOR_report << setw(5) << MTMXOR;
        if (OriXOR<MTMXOR) {
            All_XOR_report << "Ori <  MTM <= GJm\n";
            Ori_best++;
        }
        else if (MTMXOR<GJ_XOR){
            if (GJ_XOR <= OriXOR) All_XOR_report << "MTM <  GJm <= Ori\n";
            else All_XOR_report << "MTM <= Ori < GJm\n";
            MTM_best++;
        } else {
            All_XOR_report << "GJm == MTM <= Ori\n";
            GJm_best++;
        }

        MITM_XOR_report << setw(2) << MTMXOR;
        if (ele_i<((1<<fieldsize)-1)) {
            MITM_XOR_report << ", ";
        }
        if ((ele_i+1)%(fieldsize*2)==0){
            MITM_XOR_report << "\n ";
        }

        Minimal_Seq_report << " XOR seq = " << MTMXOR << "\n";
        for (unsigned int i=0; i<Stage_state.size(); i++){
            sort_V_SortNetwork/*sort_V*/(Stage_state.at(i));
            print_to_file(Stage_state[i], 2, Minimal_Seq_report);
        }

        Minimal_Seq_report << "Goal:\n";
        print_to_file(V_e, 2, Minimal_Seq_report);
        Minimal_Seq_report << "\n";

        cout << endl << endl;
        All_XOR_report << flush;
        MITM_XOR_report << flush;
        Minimal_Seq_report << flush;

    }/* Loop over requested elements */


    All_XOR_report<<"In summary, among the "<<Ori_best+MTM_best+GJm_best<<" elements.\n";
    All_XOR_report<<Ori_best<<" elements are worse even with MITM method,\n";
    All_XOR_report<<MTM_best<<" elements are improved with MITM method, and\n";
    All_XOR_report<<GJm_best<<" elements achieved minimal with Gauss-Jordan method.\n";
    All_XOR_report<<"Total XOR count = "<<TOTAL_XOR;
    All_XOR_report.close();

    MITM_XOR_report<<"};";
    MITM_XOR_report.close();

    Minimal_Seq_report.close();

    /* Stop timer */
    stop = clock();
    t = (double) (stop-start)/CLOCKS_PER_SEC;
    printf("Run time: %f\n\n", t);

    return 0;

}
